/*
 * yest
 *
 * $Id: yest.c,v 2.7.0.4 2006/12/22 18:01:58 khawkins Exp $
 *
 * Author: Kimball Hawkins
 *
 * Descripton:
 *	This program adds or subtracts days, hours, and/or minutes from a given
 *	date, and outputs the results in a designated format.
 *
 *	This is called "yest" because the default is to output yesterday's date.
 *
 *	Originally written as a shell script 13 October, 1989, it was rewritten
 *	in C in 2001 to improve  speed  on the larger day calculations, and to
 *	add weekday functionality. The only downside to this is that, at this
 *	time, it may not calculate dates past 14-Jan-2038.
 *
 * Compile:
 *	gcc -o yest yest.c
 *
 *	(Actually, this was originally designed to compile on a fairly limited
 *	compiler, GCC is not necessarily required.)
 *
 *	For Sun OS, compile with:
 *	gcc -DSUNOS -o yest yest.c
 *
 * $Log: yest.c,v $
 * Revision 2.7.0.4  2006/12/22 18:01:58  khawkins
 * Added option "--timezone".
 *
 * Revision 2.7.0.3  2006/12/22 17:34:48  khawkins
 * Modified to compile under SunOS as well. Must use "-DSUNOS".
 *
 * Revision 2.7.0.2  2006/10/04 15:04:46  khawkins
 * Corrected hr12 value for 12:xx AM, from 0 to 12.
 *
 * Revision 2.7.0.1  2006/10/03 13:41:03  khawkins
 * Minor correction for freeBSD time_t definition.
 *
 * Revision 2.7  2006/10/01 16:42:20  khawkins
 * Modified to handle strict compile warnings. Added support for date
 * "^" string modifier. Added YEST_REGION_US define.
 *
 * Revision 2.6  2006/06/08 22:02:20  khawkins
 * Minor correction to getenv("LANG") code.
 *
 * Revision 2.5  2006/05/10 14:57:35  khawkins
 * Added support for multiple input date formats. Preference is ISO.
 * Other minor improvements, see MODLOG.
 *
 * Revision 2.4  2006/01/19 22:17:48  khawkins
 * Added output options to view version, copyright and license information.
 *
 * Revision 2.3  2005/02/11 22:25:29  khawkins
 * Added/modified output fields to match date output fields in code and
 * functionality. Added coded %L, %q and %z which are not date codes.
 *
 * Revision 2.2  2004/06/17 19:50:48  khawkins
 * DST corrections. Improved compatibility with date codes.
 *
 * Revision 1.0  2001/02/07 21:22:36  khawkins
 * Initial revision
 *
 * Copyright 2004,2006 Kimball Hawkins <khawkins@acm.org>. All rights reserved.
 *
 * This program is free software; you can redistribute  it  and/or  modify  it
 * under  the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at  your  option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY  WARRANTY;  without  even  the  implied  warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public  License  for
 * more details.
 *
 * You  should  have  received  a copy of the GNU General Public License along
 * with this program; if not, write to the  Free  Software  Foundation,  Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  US
 *
 * MODIFICATION LOG
 * DATE      WHO DESCRIPTION
 * ========= === =============================================================
 * 22-Dec-06 KRH Added option --timezone.
 * 11-Oct-06 KRH Modified to compile under SunOS as well. Must use "-DSUNOS".
 * 01-Oct-06 KRH Modified code to handle strict compile warnings. Many thanks
 *		 to Guido Cavezzali for the assistance/feedback. Added support
 *		 for the date "^" string modifier.
 * 26-Sep-06 KRH Added YEST_REGION_US to make non-US compilation simpler. Bug
 *		 fix when LANG is undefined.
 * 08-Jun-06 KRH Minor correction on getenv("LANG") code.
 * 09-May-06 KRH Added support for international formats for the input date,
 *		 with ISO standard preferred. It is only US-centric if
 *		 LANG=en_US*. Added detection and acceptance of multiple input
 *		 date formats. Added support for octal escapes in format.
 * 19-Jan-06 KRH Added options to view version and license information.
 * 15-Feb-05 KRH Additional cleanup.
 * 11-Feb-05 KRH Revised and added format codes to more closely match the
 *		 equivalent date(1) format codes.
 * 28-Jul-04 KRH Added format codes %L, %q, %z.
 * 14-Jun-04 KRH Make sure user-supplied date is not DST "corrected". If user-
 *		 supplied date contains no time - ensure DST adjustments to
 *		 the final date do not alter the day of month, zero any
 *		 output time fields (we have no time data).
 * 07-Feb-01 KRH Rewritten in C. (V2.0)
 * 05-Jan-01 KRH ...and added minute adjustments as well.
 * 03-Jan-01 KRH Added support for adding/subtracting hours. Added hour,
 *               minute, time and AM/PM to formats.
 * 15-Mar-99 KRH Minor corrections to help display for clarity.
 * 30-Dec-98 KRH Added codes %b, %B, %e and %x for those who are used to
 *               date's output codes. Improved help display.
 * 22-Dec-98 KRH Made the 'th' format applicable to all appropriate numbers.
 *               Altered help to be, hopefully, clearer.
 * 21-Dec-98 KRH Added to RCS. Added formats '%C' and '%th'. Ran Y2K testing.
 * 06-Feb-98 KRH Corrected the revision.
 * 27-Aug-97 KRH Revised for simplification and readability.
 * 09-Apr-97 KRH Corrected help function.
 * 05-Dec-96 KRH Added help. Coding format adjustment.
 * 02-Nov-94 KRH Cleaned up processing, combined duplicate code.
 * 13-Oct-89 KRH Initial creation.
 * ========= === =============================================================
 */

#include <time.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

/* Define YEST_REGION_US if "en_US" is the default language. Leave undefined
 * if any other language is default. This controls the default "ambiguous
 * input" format. */
#define YEST_REGION_US

#define ONE_DAY   86400		/* Number of seconds in one day		*/
#define ONE_HOUR   3600		/* Number of seconds in one hour	*/
#define ONE_MINUTE   60		/* Number of seconds in one minute	*/

#define FMT_SIZE   2048 

typedef int bool;
#define TRUE	1
#define FALSE	0

typedef int dformat;
#define UNDEF	0
#define OTHER	1
#define ISO	2

/* ===========================================================================
 * GLOBAL VARIABLES */

char	*version="$Id: yest.c,v 2.7.0.4 2006/12/22 18:01:58 khawkins Exp $";

/* Month names and abbreviations: */
struct {
    char Name[10];
    char Abbrev[4];
} Months[] = {
	{"January",   "Jan"}, {"February",  "Feb"}, {"March",     "Mar"},
	{"April",     "Apr"}, {"May",       "May"}, {"June",      "Jun"},
	{"July",      "Jul"}, {"August",    "Aug"}, {"September", "Sep"},
	{"October",   "Oct"}, {"November",  "Nov"}, {"December",  "Dec"}
};
/* Uppercase for "%^" */
struct {
    char Name[10];
    char Abbrev[4];
} MONTHS[] = {
	{"JANUARY",   "JAN"}, {"FEBRUARY",  "FEB"}, {"MARCH",     "MAR"},
	{"APRIL",     "APR"}, {"MAY",       "MAY"}, {"JUNE",      "JUN"},
	{"JULY",      "JUL"}, {"AUGUST",    "AUG"}, {"SEPTEMBER", "SEP"},
	{"OCTOBER",   "OCT"}, {"NOVEMBER",  "NOV"}, {"DECEMBER",  "DEC"}
};

/* Days of the week and abbreviations: */
struct {
    char Name[10];
    char Abbrev[4];
} Wdays[] = {
	{"Sunday",    "Sun"}, {"Monday",    "Mon"}, {"Tuesday",   "Tue"},
	{"Wednesday", "Wed"}, {"Thursday",  "Thu"}, {"Friday",    "Fri"},
	{"Saturday",  "Sat"}
};
/* Uppercase for "%^" */
struct {
    char Name[10];
    char Abbrev[4];
} WDAYS[] = {
	{"SUNDAY",    "SUN"}, {"MONDAY",    "MON"}, {"TUESDAY",   "TUE"},
	{"WEDNESDAY", "WED"}, {"THURSDAY",  "THU"}, {"FRIDAY",    "FRI"},
	{"SATURDAY",  "SAT"}
};

char	def_date_fmt[] = { '%','x','\0' };	/* Default date format: '%x' */
char	def_time_fmt[] = { '-','%','R','\0' };	/* Default time format: '-%R'*/

char	thispgm[60],		/* Name of this program			*/
	*lang;			/* Value of the LANG env variable	*/
time_t	dfield[3];		/* Array of date fields: year, month, day
				 * (not necessarily in that order)	*/
int	tfield[2],		/* Array of time fields: hour and minute */
	rel_cent;		/* Century relative to 1900		*/
struct	tm *theTime;		/* For converting to/from epoch time	*/
bool	time_ok = TRUE,		/* time value available?		*/
	us_format = FALSE,	/* Default to the mm/dd/yyyy format?	*/
	force_upper = FALSE;	/* If "%^" prefix, force uppercase	*/

/* ===========================================================================
 * functions declarations
 */
void yest_help(void);
void yest_version(void);
void store_year(int);
void store_month(int);
void store_month_day(int);
void store_time(void);
time_t date_epoch(void);
time_t date_eu(void);
time_t date_iso(void);
time_t date_us(void);

/* ===========================================================================
 * yest_help
 */
void yest_help()
{
    char cmdline[80];
    printf("yest\n");
    printf("\n");
    printf("NAME\n");
    printf("       yest - date manipulation utility.\n");
    printf("\n");
    printf("SYNOPSIS\n");
    printf("       yest --help\n");
    printf("       yest --version\n");
    printf("       yest --license\n");
    printf("       yest [--idf={str}] [[-|+]adjust[d|h|m]] [date]");
    printf(" [format-string]\n");
    printf("\n");
    printf("DESCRIPTION\n");
    printf("       This is called \"yest\" because  the default is to");
    printf(" output yesterday´s\n");
    printf("       date.  This utility knows about leap year, daylight");
    printf(" savings time, and\n");
    printf("       such variations.\n");
    printf("\n");
    printf("       This utility adds or subtracts days, hours, and/or");
    printf(" minutes from a given\n");
    printf("       date, and outputs the results in the specified format.\n");
    printf("\n");
    printf("       The default, if no adjustment is specified, is \"-1d\"");
    printf(" (subtract one day).\n");
    printf("\n");
    printf("       Yest supports most date(1) format codes as well as a few");
    printf(" of its own.\n");
    printf("\n");
    printf("OPTIONS\n");
    printf("       --help	Display help information.\n");
    printf("\n");
    printf("       --license	Display license information.\n");
    printf("\n");
    printf("       --version	Display version information.\n");
    printf("\n");
    printf("       --input-date-format={str}\n");
    printf("       --idf={str}	Specify the default format for date input.\n");
    printf("			The {str} identifier may be:\n");
    printf("			   \"ISO\"	(Format: yyyy-mm-ddThh:mi)\n");
    printf("			   \"JP\"\t	(Format: yyyy/mm/dd-hh:mi)\n");
    printf("			   \"EU\"\t	(Format: dd/mm/yyyy-hh:mi)\n");
    printf("			   \"US\"\t	(Format: mm/dd/yyyy-hh:mi)\n");
    printf("\n");
    printf("			If the format can be determined by the");
    printf(" data, it\n");
    printf("			will be. The default format will only be");
    printf(" used for\n");
    printf("			ambiguous dates, such as \"02/03/04\".\n");
    printf("\n");
    printf("       --timezone={tz}\n");
    printf("       --tz={tz}	Specify the output date time zone. The {tz}");
    printf(" value may\n");
    printf("			be any timezone designation that your OS");
    printf(" recognizes.\n");
    printf("			If {tz} is unrecognized, the results are");
    printf(" installation\n");
    printf("			dependent. See tzselect(1).\n");
    printf("\n");
    printf("PARAMETERS\n");
    printf("       (Note that the order of the parameters is not");
    printf(" important.)\n");
    printf("\n");
    printf("       adjustment\n");
    printf("		The adjustment can be any of the following:\n");
    printf("\n");
    printf("		[-|+]day-time\n");
    printf("			Day, hour and minute adjustment");
    printf(" specification in the\n");
    printf("			format \"[-|+]day-hour:minute\". For");
    printf(" example: \"+4-12:00\".\n");
    printf("\n");
    printf("		[-|+]intd\n");
    printf("			Explicit day adjustment amount.\n");
    printf("\n");
    printf("		[-|+]inth\n");
    printf("			Explicit hour adjustment amount.\n");
    printf("\n");
    printf("		[-|+]intm\n");
    printf("			Explicit minute adjustment amount.\n");
    printf("\n");
    printf("		You  can  specify  a combination of day, hour and");
    printf(" minute\n");
    printf("		adjustments.  For example: \"yest +3d -2h\"\n");
    printf("\n");
    printf("		If no adjustment is specified, the default is");
    printf(" \"-1d\".\n");
    printf("\n");
    printf("       date	Date to be altered.  Default is today, current");
    printf(" hour, minute and\n");
    printf("		second.  {date} should be specified in ISO date");
    printf(" format:\n");
    printf("			\"yyyy-mm-dd\"\n");
    printf("		Hours and minutes may be included:\n");
    printf("			\"yyyy-mm-ddThh\"	\t(date + hour)\n");
    printf("			\"yyyy-mm-ddThh:mi\"	(date + hour and");
    printf(" minute)\n");
    printf("		\t(Note that the \"T\" before the time is a literal");
    printf(" 'T')\n");
    printf("		Other date formats recognized are:\n");
    printf("			epoch seconds	\t(seconds since 1/1/1970)\n");
    printf("			\"dd/mm/yyyy\"\n");
    printf("			or \"mm/dd/yyyy\"\t\t(if LANG=\"en_US*\")\n");
    printf("		Hours and minutes may be included as follows:\n");
    printf("			\"dd/mm/yyyy-hh\"	(date + hour)\n");
    printf("			\"dd/mm/yyyy-hh:mi\"	(date + hour and");
    printf(" minute)\n");
    printf("		If you specify a date but no time, the time");
    printf(" output fields\n");
    printf("		will, of course, be zero.  Currently, {date} must");
    printf(" be between\n");
    printf("		January 1, 1970 and January 18, 2038.\n");
    printf("\n");
    printf("       format-string\n");
    printf("		String containing date output codes and,");
    printf(" optionally, other text.\n");
    printf("		Any text in the format-str that is not a");
    printf(" recognized output code\n");
    printf("		is  passed through unchanged. This includes the");
    printf(" \"%%\" character if\n");
    printf("		it is not followed by a recognized format");
    printf(" character.  For\n");
    printf("		compatability with date(1), a leading \"+\"");
    printf(" character in format-\n");
    printf("		string will be ignored.\n");
    printf("\n");
    printf("		If there are spaces or special characters in the");
    printf(" string, it must\n");
    printf("		be quoted.  The  program  will recognize most");
    printf(" echo-like escape\n");
    printf("		conventions such as \"\\n\" (newline) and \"\\t\"");
    printf(" (tab) in format-str,\n");
    printf("		and octal escapes (\\0nn) are supported.\n");
    printf("\n");
    printf("		If only a day adjustment is specified, the default");
    printf(" format-str is\n");
    printf("		\"%%x\".  If adjustment includes a time element,");
    printf(" the default\n");
    printf("		format-str becomes \"%%x-%%R\".\n");
    printf("\n");
    printf("Adjustment specifications\n");
    printf("\n");
    printf("       Note that the sign ('+' or '-') is required, but any");
    printf(" field in the\n");
    printf("       adjustment specification is optional.  If any of the");
    printf(" numbers are\n");
    printf("       missing, they default to zero.  If the unit specification");
    printf(" is missing,\n");
    printf("       it defaults to \"d\". The program can figure out the full");
    printf(" specification\n");
    printf("       given enough syntax. For example:\n");
    printf("\n");
    printf("	+------------+------------+---------------------------------");
    printf("----------+\n");
    printf("	| ADJUSTMENT | EQUIVALENT | MEANING                         ");
    printf("          |\n");
    printf("	+------------+------------+---------------------------------");
    printf("----------+\n");
    printf("	|      -1-12 |   -1-12:00 | Subtract one day and 12 hours   ");
    printf("          |\n");
    printf("	|        +-3 |    +0-3:00 | Add three hours                 ");
    printf("          |\n");
    printf("	|        -3: |    -0-3:00 | Subtract three hours            ");
    printf("          |\n");
    printf("	|       -:30 |    -0-0:30 | Subtract 30 minutes             ");
    printf("          |\n");
    printf("	|         -2 |        -2d | Subtract two days               ");
    printf("          |\n");
    printf("	|         +h |        +0h | Add zero hours (don´t alter date");
    printf(" or time) |\n");
    printf("	+------------+------------+---------------------------------");
    printf("----------+\n");
    printf("\n");
    printf("Output format string specification\n");
    printf("    Format codes:\n");
    printf("       %%a	Abbreviated day of week (\"Tue\")\n");
    printf("       %%A	Full day of week (\"Tuesday\")\n");
    printf("       %%b	Month name abbreviated to 3 chars (\"Oct\")\n");
    printf("       %%B	Full month name (\"October\")\n");
#ifdef SUNOS
    printf("       %%c	Equivalent to \"%%a %%b %%e %%R %%Y\"\n");
#else
    printf("       %%c	Equivalent to \"%%a %%b %%e %%R %%Z %%Y\"\n");
#endif
    printf("       %%C	Century (\"19\")\n");
    printf("       %%d	Day of month (\"13\")\n");
    if ( us_format )
	printf("       %%D	Equivalent to \"%%m/%%d/%%y\"\n");
    else
	printf("       %%D	Equivalent to \"%%d/%%m/%%y\"\n");
    printf("       %%e	Equivalent to \"%%_d\" (pad with space)\n");
    printf("       %%f	Equivalent to \"%%A, %%B %%-dth, %%Y - %%I:%%M:%%S");
    printf(" %%P\"\n");
    printf("       %%F	Equivalent to \"%%Y-%%m-%%d\"\n");
    printf("       %%h	Equivalent to \"%%b\"\n");
    printf("       %%H	Hour (24 hour)\n");
    printf("       %%I	Hour (12 hour)\n");
    printf("       %%j	Day of year (\"123\")\n");
    printf("       %%k	Equivalent to \"%%_H\" (pad with space)\n");
    printf("       %%l	Equivalent to \"%%_I\" (pad with space)\n");
    printf("       %%L	Last day of month\n");
    printf("       %%m	Month number (\"10\")\n");
    printf("       %%M	Minute\n");
    printf("       %%n	Equivalent to \"\\n\" (newline).\n");
    printf("       %%p	\"AM\" or \"PM\" for the hour\n");
    printf("       %%P	\"am\" or \"pm\" for the hour\n");
    printf("       %%q	\"1\" if daylight savings time is active, \"0\"");
    printf(" if not.\n");
    printf("       %%r	Equivalent to \"%%I:%%M:%%S %%p\"\n");
    printf("       %%R	Equivalent to \"%%H:%%M\"\n");
    printf("       %%s	Number of seconds since 01/01/1970-00:00.\n");
    printf("       %%S	Seconds.\n");
    printf("       %%t	Equivalent to \"\\t\" (tab).\n");
    printf("       %%T	Equivalent to \"%%H:%%M:%%S\"\n");
    printf("       %%u	Numeric day of week: 1-7 (1=Monday).\n");
    printf("       %%w	Numeric day of week: 0-6 (0=Sunday).\n");
    if ( us_format )
	printf("       %%x	Equivalent to \"%%m/%%d/%%Y\"\n");
    else
	printf("       %%x	Equivalent to \"%%d/%%m/%%Y\"\n");
    printf("       %%X	Equivalent to \"%%H:%%M:%%S\"\n");
    printf("       %%y	Year of century (\"89\")\n");
    printf("       %%Y	Full year, including century (\"1989\")\n");
    if ( us_format )
	printf("       %%z	Equivalent to \"%%m/%%d/%%y\" (no century)\n");
    else
	printf("       %%z	Equivalent to \"%%d/%%m/%%y\" (no century)\n");
#ifndef SUNOS
    printf("       %%Z	Time zone (\"PST\")\n");
#endif
    printf("       %%%%	The \"%%\" character\n");
    printf("\n");
    printf("    Code prefixes (\"-\", \"_\"):\n");
    printf("       By  default,  yest pads numeric fields with");
    printf(" zeroes like date(1).\n");
    printf("       It also recognizes the following modifiers");
    printf(" between %% and a\n");
    printf("       numeric directive.\n");
    printf("	   '-' (hyphen) do not pad the field\n");
    printf("	   '_' (underscore) pad the field with spaces\n");
    printf("\n");
    printf("    Code prefix (\"^\"):\n");
    printf("       It also recognizes the following string modifier.\n");
    printf("       '^' (caret) use uppercase if possible\n");
    printf("\n");
    printf("    Code suffix \"th\":\n");
    printf("       Numeric format fields may be followed with the");
    printf(" modifier \"th\"\n");
    printf("       (for example: \"%%dth\"). This appends \"st\",");
    printf(" \"nd\", \"rd\", or \"th\",\n");
    printf("       as appropriate, to the number.\n");
    printf("\n");
    /* ---------------------------------------------------------------- *
     * NOTE: We generate the EXAMPLES using real time yest output.	*
     * ---------------------------------------------------------------- */
    printf("EXAMPLES\n");
    printf("       Assuming the current date and time are:\n");
    printf("	$ date\n");
    printf("	");
    fflush(stdout);
    system("date");
    printf("\n");
    printf("      Display yesterday's date in the default format.\n");
    printf("	$ yest\n");
    printf("	");
    strcpy(cmdline, thispgm);
    fflush(stdout);
    system(cmdline);
    printf("\n");
    printf("      Generate the name of the sar file for 14 days ago:\n");
    printf("	$ SARFILE=/var/adm/sa/sa`yest -14 %%d`\n");
    printf("	$ echo $SARFILE\n");
    printf("	/var/adm/sa/sa");
    strcpy(cmdline, thispgm);
    strcat(cmdline, " -14 %d");
    fflush(stdout);
    system(cmdline);
    printf("\n");
    printf("      Display the date four days ago with a specified output");
    printf(" format.\n");
    printf("	$ yest -4 \"%%b %%e %%Y\"\n");
    printf("	");
    strcpy(cmdline, thispgm);
    strcat(cmdline, " -4 '%b %e %Y'");
    fflush(stdout);
    system(cmdline);
    printf("\n");
    printf("      Calculate the date an arbitrary number of days from a");
    printf(" specified date and\n");
    printf("      display it with an extensive format:\n");
    printf("	$ COUNT=+4\n");
    printf("	$ yest \"The date is:\\n\\tThe %%dth day of %%B, %%Y\"");
    printf(" 1988-12-28 $COUNT\n");
    printf("	The date is:\n");
    printf("		The 1st day of January, 1989\n");
    printf("\n");
    printf("      Display the date and time (default format) of three");
    printf(" hours ago:\n");
    printf("	$ yest -3h\n");
    printf("	");
    strcpy(cmdline, thispgm);
    strcat(cmdline, " -3h");
    fflush(stdout);
    system(cmdline);
    printf("\n");
    printf("AUTHOR\n");
    printf("       Kimball Hawkins\n");
    printf("\n");
    exit(0);
}

/* ===========================================================================
 * yest_version
 */
void yest_version()
{
    printf("yest\n");
    printf("	%s\n", "$Revision: 2.7.0.4 $");
    printf("	%s\n", "$Author: khawkins $");
    printf("	%s\n", "$Date: 2006/12/22 18:01:58 $");
    printf("\n");
    printf("Copyright 2004,2006 Kimball Hawkins <khawkins@acm.org>.");
    printf(" All rights reserved.\n");
    printf("\n");
    printf("This program is free software; you can redistribute  it  and/or");
    printf("  modify  it\n");
    printf("under  the terms of the GNU General Public License as published");
    printf(" by the Free\n");
    printf("Software Foundation; either version 2 of the License, or (at ");
    printf(" your  option)\n");
    printf("any later version.\n");
    printf("\n");
    printf("This program is distributed in the hope that it will be useful,");
    printf(" but WITHOUT\n");
    printf("ANY  WARRANTY;  without  even  the  implied  warranty of");
    printf(" MERCHANTABILITY or\n");
    printf("FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public ");
    printf(" License  for\n");
    printf("more details.\n");
    printf("\n");
    printf("You  should  have  received  a copy of the GNU General Public");
    printf(" License along\n");
    printf("with this program; if not, write to the  Free  Software ");
    printf(" Foundation,  Inc.,\n");
    printf("59 Temple Place, Suite 330, Boston, MA  02111-1307  US\n");
    printf("\n");
    exit(0);
}

/* ======================================================================== 
 * DATE FIELD STORAGE
 * ======================================================================== */

/* ------------------------------------------------------------------------ 
 * store_year
 * Allow for 1, 2 or 4 digit year
 */
void store_year(idx)
int idx;
{
    if ( dfield[idx] < 100 ) {
	if ( dfield[idx] > 37 ) {
	    /* Assume it is last century */
	    theTime->tm_year = dfield[idx] + rel_cent - 100;
	} else {
	    /* Assume it is this century */
	    theTime->tm_year = dfield[idx] + rel_cent;
	}
    } else {
	theTime->tm_year = dfield[idx];
    }
    /* If the full year was entered, make it relative */
    if ( theTime->tm_year > 1900 )
	theTime->tm_year -= 1900;
}

/* ------------------------------------------------------------------------ 
 * store_month
 */
void store_month(idx)
int idx;
{
    if ( dfield[idx] > 12 ) {
	fprintf(stderr, "ERROR: Bad input month: '%ld'\n", (long) dfield[idx]);
	exit(1);
    }
    theTime->tm_mon = dfield[idx] - 1;
}

/* ------------------------------------------------------------------------ 
 * store_month_day
 */
void store_month_day(idx)
int idx;
{
    if ( dfield[idx] > 31 ) {
	fprintf(stderr, "ERROR: Bad input day of month: '%ld'\n",
		(long) dfield[idx]);
	exit(1);
    }
    theTime->tm_mday = dfield[idx];
}

/* ------------------------------------------------------------------------ 
 * store_time
 */
void store_time()
{
    if ( time_ok == FALSE ) return;	/* No time information, skip it */

    /* Hour */
    if ( tfield[0] > 24 ) {
	fprintf(stderr, "ERROR: Bad input hour: '%d'\n", tfield[0]);
	exit(1);
    }
    theTime->tm_hour = tfield[0];

    /* Minute */
    if ( tfield[1] > 0 ) {
	if ( tfield[1] > 60 ) {
	    fprintf(stderr, "ERROR: Bad input minute: '%d'\n", tfield[1]);
	    exit(1);
	}
	theTime->tm_min = tfield[1];
    }
}

/* ======================================================================== 
 * HANDLING FOR DIFFERENT DATE FORMATS
 * ======================================================================== */

/* ------------------------------------------------------------------------ 
 * date_epoch
 *	Just one numeric field
 */
time_t date_epoch()
{
    /* Nothing to do, just return that one field */
    return( dfield[0] );
}

/* ------------------------------------------------------------------------ 
 * date_eu
 *	Fields: DD	MM	YYYY	[HH	MI]
 */
time_t date_eu()
{
    store_month_day(0);	/* Store dfield 0 in day of month */
    store_month(1);	/* Store dfield 1 in month	*/
    store_year(2);	/* Store dfield 2 in year	*/
    store_time();	/* Store hour/minutes if any	*/

    return( mktime(theTime) );
}

/* ------------------------------------------------------------------------ 
 * date_iso
 *	Fields: YYYY	MM	DD	[HH	MI]
 */
time_t date_iso()
{
    store_year(0);	/* Store dfield 0 in year	*/
    store_month(1);	/* Store dfield 1 in month	*/
    store_month_day(2);	/* Store dfield 2 in day of month */
    store_time();	/* Store hour/minutes if any	*/

    return( mktime(theTime) );
}

/* ------------------------------------------------------------------------ 
 * date_us
 *	Fields: MM	DD	YYYY	[HH	MI]
 */
time_t date_us()
{
    store_month(0);	/* Store dfield 0 in month	*/
    store_month_day(1);	/* Store dfield 1 in day of month */
    store_year(2);	/* Store dfield 2 in year	*/
    store_time();	/* Store hour/minutes if any	*/

    return( mktime(theTime) );
}

/* ========================================================================
 * MAIN
 * ======================================================================== */

int main(int argc, char *argv[])
{
    /* ---------------------------------------------------------------------
     * INITIALIZATION
     */
    int		i, j, k, temp,
		d, t,			/* Date and time field indexes	*/
		lastday,		/* Last day of month		*/
		hr12,			/* hour in 12 hour format	*/
		prior_isdst,		/* tm_isdst before mktime() call*/
		tot_adjust = 0;		/* Total adjustment in seconds	*/
    bool	a_specified = FALSE,	/* Adjustment specified?	*/
		d_specified = FALSE,	/* Base date specified?		*/
		f_specified = FALSE;	/* Format specified?		*/
    char	temp_str[20],
		pad_char,		/* Padding character		*/
		Dth[3],			/* "th" value for Day		*/
		mth[3],			/* "th" value for Month		*/
		Yth[3],			/* "th" value for Year		*/
		Hth[3],			/* "th" value for Hour24	*/
		Ith[3],			/* "th" value for Hour12	*/
		jth[3],			/* "th" value for Day of year	*/
		lth[3],			/* "th" value for Last DoM	*/
		Mth[3],			/* "th" value for Minute	*/
		Sth[3],			/* "th" value for Second	*/
		wth[3],			/* "th" value for Day of week	*/
		AM[3],			/* "AM" or "PM"			*/
		AMlc[3],		/* "am" or "pm"			*/
		format[FMT_SIZE],	/* Date output format		*/
		exp_fmt[FMT_SIZE],	/* Format with equivs expanded	*/
		sign,			/* Adjustment sign (+,-)	*/
		time_fmt[4],		/* Time only output format	*/
		timezone[32],		/* Possible output timezone	*/
		type;			/* Type of input		*/
    dformat	date_format = UNDEF;	/* Will be either ISO or OTHER	*/
    time_t	unixtime, savetime,
		(*def_date) (void);	/* Pointer to a date-store function */

    /* --------------------------------------------------------------------- *
     * Initialize variables */
    time_fmt[0] = '\0';
    timezone[0] = '\0';
    strcpy(thispgm, argv[0]);

    unixtime = time(0);
    theTime = localtime(&unixtime);
    prior_isdst = theTime->tm_isdst;	/* Save DST flag for later */

    /* Default century: Relative - how many hundreds of years since 1900 */
    rel_cent  = (theTime->tm_year / 100) * 100;

    for ( i = 0; i < 3; i++ ) dfield[i] = 0;
    for ( i = 0; i < 2; i++ ) tfield[i] = 0;

    /* Determine if the default date format is US or not: */
    lang = (char *) getenv("LANG");
#ifdef YEST_REGION_US
    /* Default is "US" */
    if ( lang == 0 || strncmp(lang, "en_US", 5) == 0 || strlen(lang) == 0 )
	us_format = TRUE;
#else
    /* Set as "US" only if explicit via LANG */
    if ( strncmp(lang, "en_US", 5) == 0 ) us_format = TRUE;
#endif

    if ( lang != 0 && strncmp(lang, "en_US", 5) == 0 ) us_format = TRUE;

    /* Accept input date syntax specification */
    def_date = 0;	/* No default yet */

    /* --------------------------------------------------------------------- *
     * Parse the arguments */
    for ( i = 1; i < argc; i++ ) {
	/* ----------------------------------------------------------------- *
	 * OPTIONS */
	if ( strncmp( argv[i], "--help", 6 ) == 0 ||
	     strncmp( argv[i], "-?",     2 ) == 0 ||
	     strncmp( argv[i], "?",      1 ) == 0 ||
	     strncmp( argv[i], "help",   4 ) == 0 )
	    yest_help();	/* help requested, display it */

	if ( strncmp( argv[i], "--version", 9 ) == 0 ||
	     strncmp( argv[i], "--license", 9 ) == 0 )
	    yest_version();	/* version/license information requested */

	if ( strncmp( argv[i], "--timezone=", 11 ) == 0 ||
	     strncmp( argv[i], "--tz=", 5 ) == 0 ) {
	    /* User-defined output time zone */

	    /* Get the location of the '=' character */
	    for ( j = 0; j < strlen(argv[i]); j++ )
		if ( argv[i][j] == '=' ) break;

	    /* Store the value (whatever it is) */
	    strncpy( timezone, &argv[i][++j], 31 );

	    continue;
	}
	if ( strncmp( argv[i], "--input-date-format=", 20 ) == 0 ||
	     strncmp( argv[i], "--idf=", 6 ) == 0 ) {
	    /* User-defined input date format */

	    /* Get the location of the '=' character */
	    for ( j = 0; j < strlen(argv[i]); j++ )
		if ( argv[i][j] == '=' ) break;

	    /* Look at the next character */
	    switch ( argv[i][++j] ) {
	    case 'e':
	    case 'E':
		def_date = &date_eu;
		break;
	    case 'j':
	    case 'J':
		def_date = &date_iso;	/* Same field order as ISO */
		break;
	    case 'u':
	    case 'U':
		def_date = &date_us;
		break;
	    default:
		def_date = &date_iso;
		break;
	    }
	    continue;
	}

	/* ----------------------------------------------------------------- *
	 * PARAMETERS */

	/* Determine what each parameter is: adjustment, date or format */
	type = '-';
	/* Look at the content of the string */
	for ( j = 0; j < strlen(argv[i]); j++ ) {
	    if ( argv[i][j] == '%' ) {
		/* If it contains '%' it must be a format specification */
		type = 'f';
		break;
	    }
	}
	/* If not a format specification... */
	if ( type == '-' ) {
	    if ( argv[i][0] == '-' || argv[i][0] == '+' )
		type = 'a';		/* Is adjustment */
	    else
		type = 'd';		/* Default is date*/
	}

	/* Handle each type of parameter */
	switch (type) {
	case 'd':
		/* -----------------------------------------------------------
		 * DATE */
		d_specified = TRUE;

		time_ok = FALSE;	/* No time available, yet */

		/* Initialize the time struct */
		theTime->tm_sec  = theTime->tm_min = 0;
		theTime->tm_mday = theTime->tm_mon = theTime->tm_year = 0;
		theTime->tm_hour = 12;	/* To allow for DST adjustments */

		/* Parse out the various date/time fields */
		for ( j = 0, d = 0; j < strlen(argv[i]); j++ ) {
		    if ( argv[i][j] == '-' ) {
			/* '-' is the ISO separator */
			date_format = ISO;

		    } else if ( argv[i][j] == '/' ) {
			/* The '/' separator is non-ISO format */
			date_format = OTHER;

		    } else if (  argv[i][j] == 'T' ) {
			/* ISO time specifier */
			if ( d < 2  ||  date_format == OTHER ) {
			    /* Must have all 3 date fields at this point */
			    fprintf(stderr, "ERROR: Bad input date: '%s'\n",
					argv[i]);
			    exit(1);
			}
			break;		/* Break to extract time */

		    } else if ( ! isdigit((unsigned char) argv[i][j]) ) {
			fprintf(stderr, "ERROR: Bad input date: '%s'\n",
				argv[i]);
			fprintf(stderr, "       Bad digit '%c'.\n",
				argv[i][j]);
			exit(1);

		    } else if ( d > 2 ) {
			/* Any remaining fields would be time info. */
			break;

		    } else {
			/* As long as char is a digit, store in temp_str */
			strcpy(temp_str, "0");
			for ( k = 0; isdigit((unsigned char) argv[i][j]); k++ )
			    temp_str[k] = argv[i][j++];
			/* Convert to integer */
			temp_str[k] = '\0';
			dfield[d++] = atoi(temp_str);
		    }
		}

		/* Extract time if any string left */
		for ( t = 0; j < strlen(argv[i]); j++ ) {
		    if ( argv[i][j] == ':' ) {
			continue;

		    } else if ( ! isdigit((unsigned char) argv[i][j]) ) {
			fprintf(stderr, "ERROR: Bad input date: '%s'\n",
				argv[i]);
			fprintf(stderr, "       Bad digit '%c'.\n",
				argv[i][j]);
			exit(1);

		    } else if ( t > 1 ) {
			/* Ignore any seconds field */
			fprintf(stderr,
				"WARNING: Ignoring excess time data in '%s'\n",
				argv[i]);
			break;

		    } else {
			/* As long as char is a digit, store in temp_str */
			strcpy(temp_str, "0");
			for ( k = 0; isdigit((unsigned char) argv[i][j]); k++ )
			    temp_str[k] = argv[i][j++];
			/* Convert to integer */
			temp_str[k] = '\0';
			tfield[t++] = atoi(temp_str);
			time_ok = TRUE;
		    }
		}

		/* If no default format determined yet, figure it out */
		if ( def_date == 0 ) {
		    if ( date_format == ISO )
			def_date = &date_iso;

		    else if ( us_format )
			def_date = &date_us;

		    else
			def_date = &date_eu;
		}

		/* Convert the raw data into standard time */
		switch (d) {	
		case 1:
		    /* Only one field - must be epoch */
		    unixtime = date_epoch();
		    break;
		case 3:
		    /* We have year, month and day, but don't know in what
		     * order. Figure it out, if we can. */
		    if ( dfield[0] > 31 ) {
			/* Treat as YYYY-MM-DD */
			unixtime = date_iso();

		    } else if ( dfield[0] > 12 && dfield[0] < 32 &&
				dfield[1] < 13 && dfield[2] > 31 ) {
			/* Must be dd/mm/yyyy */
			unixtime = date_eu();

		    } else if ( dfield[0] < 13 && dfield[2] > 31 &&
				dfield[1] > 12 && dfield[1] < 32 ) {
			/* Must be mm/dd/yyyy */
			unixtime = date_us();

		    } else {
			/* Can't guess, use the "default" */
			unixtime = def_date();
		    }
		    break;
		default:
		    /* Should already have been detected */
		    fprintf(stderr, "ERROR: Invalid input date: '%s'\n",
			argv[i]);
		    exit(1);
		}

		if ( unixtime == -1 ) {
		    /* Date could not be converted */
		    fprintf(stderr, "ERROR: Date out of range: '%s'\n",
			argv[i]);
		    fprintf(stderr, "       See \"yest --help\" for help\n");
		    exit(1);
		}
		/*
		 * Daylight Savings Time fix:
		 * If the prior tm_isdst does not match the new tm_isdst, then
		 * mktime has just automatically adjusted the date/time for us.
		 * Since this is a user-specified, absolute date/time, this is
		 * NOT what we want.
		 */
		theTime->tm_hour += ( prior_isdst - theTime->tm_isdst );
		/* Re-execute the mktime call if hour is now out of range */
		if ( theTime->tm_hour > 23 || theTime->tm_hour < 0 )
		    unixtime = mktime(theTime);
		break;

	case 'f':
		/* -----------------------------------------------------------
		 * FORMAT */
		f_specified = TRUE;

		if ( argv[i][0] == '+' ) {
		    strcpy(format, ++argv[i]);	/* Date "+" format intro */
		} else {
		    strcpy(format, argv[i]);
		}
		break;

	default:
		/* -----------------------------------------------------------
		 * ADJUSTMENT - May be day, hour and/or minute
		 *	Explicit: [+-]{n}[dhm],
		 *	Context:  [+-]{d}-{h}:{m}
		 */
		a_specified = TRUE;

		j = 0;
		sign = '-'; /* If missing, the sign is negative */
		switch (argv[i][0]) {
		case '-':	sign = '-'; j++; break;
		case '+':	sign = '+'; j++; break;
		}
		temp = 0;
		/* Get the numeric portion of argument */
		while ( isdigit((unsigned char) argv[i][j]) )
		    temp = (temp * 10) + (argv[i][j++] - 48);

		switch (argv[i][j]) {
		case '\0':
			/* Implicit day specification */
		case 'd':
			/* Explicit day specification */
			if ( sign == '-' )
			    tot_adjust = tot_adjust - ( temp * ONE_DAY );
			else
			    tot_adjust = tot_adjust + ( temp * ONE_DAY );
			j++;
			break;
		case 'h':
			/* Explicit hour specification */
			if ( sign == '-' )
			    tot_adjust = tot_adjust - ( temp * ONE_HOUR );
			else
			    tot_adjust = tot_adjust + ( temp * ONE_HOUR );
			j++;
			strcpy(time_fmt, def_time_fmt);
			break;
		case 'm':
			/* Explicit minute specification */
			if ( sign == '-' )
			    tot_adjust = tot_adjust - ( temp * ONE_MINUTE );
			else
			    tot_adjust = tot_adjust + ( temp * ONE_MINUTE );
			j++;
			strcpy(time_fmt, def_time_fmt);
			break;
		case '-':
			/* Context - date/time format (i.e. "-3-12:30") */
			if ( sign == '-' )
			    tot_adjust = tot_adjust - ( temp * ONE_DAY );
			else
			    tot_adjust = tot_adjust + ( temp * ONE_DAY );

			j++;
			strcpy(time_fmt, def_time_fmt);
			temp = 0;
			/* Get the hour */
			while ( isdigit((unsigned char) argv[i][j]) )
			    temp = (temp * 10) + (argv[i][j++] - 48);
			/* -------------------------------------------- *
			 *  NOTE: This code drops through on purpose	*
			 * -------------------------------------------- */
		case ':':
			/* Time portion of context format (i.e. "-12:34") */
			if ( sign == '-' )
			    tot_adjust = tot_adjust - ( temp * ONE_HOUR );
			else
			    tot_adjust = tot_adjust + ( temp * ONE_HOUR );

			if ( argv[i][j] == ':' ) {
			    temp = 0;
			    /* Get the minute */
			    while ( isdigit((unsigned char) argv[i][++j]) )
				temp = (temp * 10) + (argv[i][j] - 48);
			    if ( sign == '-' )
				tot_adjust = tot_adjust - ( temp * ONE_MINUTE );
			    else
				tot_adjust = tot_adjust + ( temp * ONE_MINUTE );
			}
			if ( argv[i][j] != '\0' ) {
			    /* Invalid adjustment */
			    fprintf(stderr,
				"ERROR: Invalid adjustment: '%s'\n", argv[i]);
			    fprintf(stderr,
				"       See \"yest --help\" for help\n");
			    exit(1);
			}
			strcpy(time_fmt, def_time_fmt);
			break;
		default:
			/* Not an adjustment, unrecognized argument */
			fprintf(stderr, "ERROR: Unrecognized argument: '%s'\n",
				argv[i]);
			fprintf(stderr,
				"       See \"yest --help\" for help\n");
			exit(1);
		}
	}
    }

    /* Add defaults for anything not explicitly specified */
    if ( ! a_specified ) tot_adjust = 0 - ONE_DAY;
    if ( ! d_specified ) unixtime = time(0);	/* Current date and time */
    if ( ! f_specified ) {
	strcpy( format, def_date_fmt);
	strcat( format, time_fmt);
    }

    /* ---------------------------------------------------------------------
     * Calculate the new date
     */

    /* If entered, change TZ to the specified value. */
    if ( timezone[0] != '\0' )
    	setenv("TZ", timezone, 1);

    unixtime += tot_adjust;
    theTime = localtime(&unixtime);

    /* NOTE:	This adjustment MAY include an additional, automatic
     *		adjustment for daylight savings time. While this may be
     *		unexpected, it is a functionality we do desire.
     */

    /* ---------------------------------------------------------------------
     * PREPARATION:
     * Get ready for any format, then process the format mask.
     */

    /* Find last day of month */
    savetime = unixtime;
    theTime->tm_mon += 1;
    theTime->tm_mday = 0;
    unixtime = mktime(theTime);
    lastday = theTime->tm_mday;
    /* Restore calculated time */
    unixtime = savetime;
    theTime = localtime(&unixtime);

    /* Calculate 12-hour time */
    if ( theTime->tm_hour == 0 )
	hr12 = 12;
    else if ( theTime->tm_hour < 13 )
	hr12 = theTime->tm_hour;
    else
	hr12 = theTime->tm_hour - 12;

    /* Figure AM/PM */
    if ( theTime->tm_hour < 12 ) {
	strcpy(AM, "AM");
	strcpy(AMlc, "am");
    } else {
	strcpy(AM, "PM");
	strcpy(AMlc, "pm");
    }

    /* Determine the format for 'th' codes: */
    switch (theTime->tm_mday) {
    case 11: case 12: case 13:
		strcpy(Dth, "th");
		break;
    default:	temp = theTime->tm_mday % 10;
		switch (temp) {
		case 1:  strcpy(Dth, "st"); break;
		case 2:  strcpy(Dth, "nd"); break;
		case 3:  strcpy(Dth, "rd"); break;
		default: strcpy(Dth, "th");
		}
    }
    switch (theTime->tm_mon) {
    case 0:	strcpy(mth, "st"); break;	/* First month */
    case 1:	strcpy(mth, "nd"); break;	/* Second month */
    case 2:	strcpy(mth, "rd"); break;	/* Third month */
    default:	strcpy(mth, "th");
    }
    temp = theTime->tm_year % 100;
    switch (temp) {
    case 11: case 12: case 13:
		strcpy(Yth, "th");
		break;
    default:	temp = temp % 10;
		switch (temp) {
		case 1:  strcpy(Yth, "st"); break;
		case 2:  strcpy(Yth, "nd"); break;
		case 3:  strcpy(Yth, "rd"); break;
		default: strcpy(Yth, "th");
		}
    }
    switch (theTime->tm_hour) {
    case 11: case 12: case 13:
		strcpy(Hth, "th");
		break;
    default:	temp = theTime->tm_hour % 10;
		switch (temp) {
		case 1:  strcpy(Hth, "st"); break;
		case 2:  strcpy(Hth, "nd"); break;
		case 3:  strcpy(Hth, "rd"); break;
		default: strcpy(Hth, "th");
		}
    }
    switch (hr12) {
    case 11: case 12: case 13:
		strcpy(Ith, "th");
		break;
    default:	temp = hr12 % 10;
		switch (temp) {
		case 1:  strcpy(Ith, "st"); break;
		case 2:  strcpy(Ith, "nd"); break;
		case 3:  strcpy(Ith, "rd"); break;
		default: strcpy(Ith, "th");
		}
    }
    switch (theTime->tm_yday % 100) {
    case 11: case 12: case 13:
		strcpy(jth, "th");
		break;
    default:	switch (theTime->tm_yday % 10) {
		case 1:  strcpy(jth, "st"); break;
		case 2:  strcpy(jth, "nd"); break;
		case 3:  strcpy(jth, "rd"); break;
		default: strcpy(jth, "th");
		}
    }
    switch (theTime->tm_min) {
    case 11: case 12: case 13:
		strcpy(Mth, "th");
		break;
    default:	temp = theTime->tm_min % 10;
		switch (temp) {
		case 1:  strcpy(Mth, "st"); break;
		case 2:  strcpy(Mth, "nd"); break;
		case 3:  strcpy(Mth, "rd"); break;
		default: strcpy(Mth, "th");
		}
    }
    switch (theTime->tm_sec) {
    case 11: case 12: case 13:
		strcpy(Sth, "th");
		break;
    default:	temp = theTime->tm_sec % 10;
		switch (temp) {
		case 1:  strcpy(Sth, "st"); break;
		case 2:  strcpy(Sth, "nd"); break;
		case 3:  strcpy(Sth, "rd"); break;
		default: strcpy(Sth, "th");
		}
    }
    switch (theTime->tm_wday) {
    case 1:	strcpy(wth, "st"); break;
    case 2:	strcpy(wth, "nd"); break;
    case 3:	strcpy(wth, "rd"); break;
    default:	strcpy(wth, "th");
    }
    switch (lastday) {
    case 11: case 12: case 13:
		strcpy(lth, "th");
		break;
    default:	temp = lastday % 10;
		switch (temp) {
		case 1:  strcpy(lth, "st"); break;
		case 2:  strcpy(lth, "nd"); break;
		case 3:  strcpy(lth, "rd"); break;
		default: strcpy(lth, "th");
		}
    }

    /* Pre-process equivalent codes */
    for ( i = 0, j = 0; i < strlen(format); ) {
	exp_fmt[j++] = format[i];
	if ( format[i++] == '%' ) {
	    exp_fmt[j] = '\0'; /* Terminate, so we can use strcat() */
	    switch (format[i]) {
#ifdef SUNOS
	    case 'c':	strcat(exp_fmt, "a %b %e %H:%M:%S %Y");
#else
	    case 'c':	strcat(exp_fmt, "a %b %e %H:%M:%S %Z %Y");
#endif
			break;
	    case 'f':	strcat(exp_fmt, "A, %B %-dth, %Y - %I:%M:%S %P");
			break;
	    case 'F':	strcat(exp_fmt, "Y-%m-%d");
			break;
	    case 'h':	strcat(exp_fmt, "b");
			break;
	    case 'n':	/* 'date' version of '\n' */
			exp_fmt[--j] = '\0';	/* back up one character */
			strcat(exp_fmt, "\\n");
			break;
	    case 'r':	strcat(exp_fmt, "I:%M:%S %p");
			break;
	    case 'R':	strcat(exp_fmt, "H:%M");
			break;
	    case 't':	/* 'date' version of '\t' */
			exp_fmt[--j] = '\0';	/* back up one character */
			strcat(exp_fmt, "\\t");
			break;
	    case 'T':	strcat(exp_fmt, "H:%M:%S");
			break;
	    case 'D':	/* Same as 'x' */
	    case 'x':	if ( us_format )
			    strcat(exp_fmt, "m/%d/%Y");
			else
			    strcat(exp_fmt, "d/%m/%Y");
			break;
	    case 'X':	strcat(exp_fmt, "H:%M:%S");
			break;
	    case 'z':	if ( us_format )
			    strcat(exp_fmt, "m/%d/%y");
			else
			    strcat(exp_fmt, "d/%m/%y");
			break;
	    default:	exp_fmt[j++] = format[i];
			exp_fmt[j] = '\0';
	    }
	    j = strlen(exp_fmt);	/* Get the new length */
	    i++;
	}
    }

    /* ---------------------------------------------------------------------
     * OUTPUT:
     * Process the format mask and output the results to standard output
     */
    for ( i = 0; i < strlen(exp_fmt); ) {
	while ( exp_fmt[i] == '%' ) {
	    /* Default: pad with "0" */
	    pad_char = '0';
	    force_upper = FALSE;
	    switch ( exp_fmt[++i] ) {
	    case '_':	pad_char = ' ';  i++; break;
	    case '-':	pad_char = '\0'; i++; break;
	    case '^':	force_upper = TRUE; i++; break;
	    case '#':	i++; break;	/* date "opposite case", ignore */
	    }

	    switch (exp_fmt[i]) {
	    case 'A':	if ( force_upper )
			    printf("%s", WDAYS[theTime->tm_wday].Name);
			else
			    printf("%s", Wdays[theTime->tm_wday].Name);
			break;
	    case 'a':	if ( force_upper )
			    printf("%s", WDAYS[theTime->tm_wday].Abbrev);
			else
			    printf("%s", Wdays[theTime->tm_wday].Abbrev);
			break;
	    case 'B':	if ( force_upper )
			    printf("%s", MONTHS[theTime->tm_mon].Name);
			else
			    printf("%s", Months[theTime->tm_mon].Name);
			break;
	    case 'b':	if (force_upper )
			    printf("%s", MONTHS[theTime->tm_mon].Abbrev);
			else
			    printf("%s", Months[theTime->tm_mon].Abbrev);
			break;
	    case 'C':	printf("%d", (theTime->tm_year / 100) + 19);
			break;
	    case 'd':	if ( theTime->tm_mday < 10 ) printf("%c", pad_char);
			printf("%d", theTime->tm_mday);
			break;
	    case 'e':	if ( theTime->tm_mday < 10 ) printf(" ");
			printf("%d", theTime->tm_mday);
			break;
	    case 'H':	if ( ! time_ok ) {
			    printf("00");
			} else {
			    if ( theTime->tm_hour < 10 ) printf("%c", pad_char);
			    printf("%d", theTime->tm_hour);
			}
			break;
	    case 'I':	if ( ! time_ok ) {
			    printf("00");
			} else {
			    if ( hr12 < 10 ) printf("%c", pad_char);
			    printf("%d", hr12);
			}
			break;
	    case 'j':	if ( theTime->tm_yday < 100 ) printf("%c", pad_char);
			if ( theTime->tm_yday < 10  ) printf("%c", pad_char);
			printf("%d", theTime->tm_yday);
			break;
	    case 'k':	if ( ! time_ok ) {
			    printf("00");
			} else {
			    if ( theTime->tm_hour < 10 ) printf(" ");
			    printf("%d", theTime->tm_hour);
			}
			break;
	    case 'l':	if ( ! time_ok ) {
			    printf("00");
			} else {
			    if ( hr12 < 10 ) printf(" ");
			    printf("%d", hr12);
			}
			break;
	    case 'L':	if ( lastday < 10 ) printf("%c", pad_char);
			printf("%d", lastday);
			break;
	    case 'M':	if ( ! time_ok ) {
			    printf("00");
			} else {
			    if ( theTime->tm_min  < 10 ) printf("%c", pad_char);
			    printf("%d", theTime->tm_min);
			}
			break;
	    case 'm':	if ( theTime->tm_mon < 9 ) printf("%c", pad_char);
			printf("%d", theTime->tm_mon + 1);
			break;
	    case 'P':	printf("%s", AMlc);
			break;
	    case 'p':	printf("%s", AM);
			break;
	    case 'q':	printf("%d", theTime->tm_isdst);
			break;
	    case 's':	printf("%ld", (long)unixtime);
			break;
	    case 'S':	if ( ! time_ok ) {
			    printf("00");
			} else {
			    if ( theTime->tm_sec < 10 ) printf("%c", pad_char);
			    printf("%d", theTime->tm_sec);
			}
			break;
	    case 'u':	if ( theTime->tm_wday == 0 )
			    printf("7");
			else
			    printf("%d", theTime->tm_wday);
			break;
	    case 'w':	printf("%d", theTime->tm_wday);
			break;
	    case 'Y':	printf("%d", theTime->tm_year + 1900);
			break;
	    case 'y':	if ( theTime->tm_year % 100 < 10 )
			    printf("%c", pad_char);
			printf("%d", theTime->tm_year % 100);
			break;
#ifndef SUNOS
	    case 'Z':	printf("%s", theTime->tm_zone);
			break;
#endif
	    case '%':	printf("%c", '%');
			break;
	    default:	/* Matched no known pattern, just print it */
			printf("%%%c", exp_fmt[i]);
	    }
	    i++;
	    /* Check the following 2 chars, process "th" suffixes */
	    if ( strncmp(&exp_fmt[i], "th", 2) == 0 ) {
		switch (exp_fmt[i-1]) {
		case 'd':	printf("%s", Dth); break;
		case 'H':	if ( time_ok ) printf("%s", Hth); break;
		case 'h':	if ( time_ok ) printf("%s", Ith); break;
		case 'j':	printf("%s", jth); break;
		case 'l':	printf("%s", lth); break;
		case 'M':	if ( time_ok ) printf("%s", Mth); break;
		case 'm':	printf("%s", mth); break;
		case 'S':	if ( time_ok ) printf("%s", Sth); break;
		case 'w':	printf("%s", wth); break;
		case 'Y':	printf("%s", Yth); break;
		case 'y':	printf("%s", Yth); break;
		default:	/* Not applicable, just print it */
				printf("th");
		}
		i += 2;
	    }
	}
	if ( i < strlen(exp_fmt) ) {
	    if ( exp_fmt[i] == '\\' ) {
		/* Process escape sequences */
		switch (exp_fmt[++i]) {
		case 't':	printf("\t"); break;
		case 'n':	printf("\n"); break;
		case 'v':	printf("\v"); break;
		case 'b':	printf("\b"); break;
		case 'r':	printf("\r"); break;
		case 'f':	printf("\f"); break;
		case '\\':	printf("\\"); break;
		case '\'':	printf("'");  break;
		case '"':	printf("\""); break;
		case '0':	/* Support for octal escapes */
				temp_str[0] = '0';
				for ( j = 1;
				      isdigit((unsigned char) exp_fmt[i+1]);
				      j++ )
				    temp_str[j] = exp_fmt[++i];
				temp_str[j] = '\0';
				printf("%c", atoi(temp_str));
				break;
		default:	printf("\\%c", exp_fmt[i]); break;
		}
		i++;
	    } else
		printf("%c", exp_fmt[i++]);
	}
    }
    /* Done */
    printf("\n");
    exit (0);
}

/*
 * End of yest
 */
